#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

import threading
from harkmw.logic.node import Node
from harkmw.logger import logger

import os
debug = (bool(os.environ['HARK_DEBUG'])
         ) if 'HARK_DEBUG' in os.environ else False


class Network(Node):
    def __init__(self):
        Node.__init__(self)
        self.name = None
        self.type = None
        self.network_type = None
        self.topic = None
        self.nodes = {}
        self.links = []
        self.inputs = {}  # NetTerminal
        self.outputs = {}  # NetTerminal
        self.conditions = {}  # NetTerminal
        self.input_terminals = {}
        self.output_terminals = {}
        self.parameters = {}

        self.cv = threading.Condition()
        self.remote_data_offset = 0
        self.remote_data = []
        self.mqtt_proxies = []
        self.mqtt_messages = []

        self.input_collector = None
        self.output_collector = None
        self.condition_collector = None

    def run(self, process):
        """
        """
        try:
            if self.has_remote_nodes(
                    process) and self.process_name == process.process_name:
                process.publish('_RUN', 0, True)

            self.proceed(0, process)
            if self.process_name == process.process_name:
                self.output_collector.dump_input(0)

        except Exception as ex:
            logger.error('Error in network.proceed(): {}'.format(str(ex)))
            raise ex

        finally:
            if self.has_remote_nodes(
                    process) and self.process_name == process.process_name:
                t = '_TERMINATE'
                process.publish(t, 0, True)

    def proceed(self, count, process, input=None):
        logger.debug(
            "Network.proceed: name={}, network_type={}, count={}, input={}".format(
                self.name, self.network_type, count, input))

        # need_publish = self.has_remote_nodes(process)
        # if self.process_name == process.process_name:
        #     if process.is_mqtt_initialized() :
        #         if need_publish :
        #             process.publish(self.get_topic() + "/_INPUT", count, True)
        execute_remote = (self.process_name !=
                          process.process_name and self.name != 'MAIN')
        is_iterator = (self.network_type == 'iterator')
        logger.debug(
            'Network.proceed: name={}, self.process_name={}, process.process_name={}: execute_remote={}, is_iterator={}'.format(
                self.name,
                self.process_name,
                process.process_name,
                execute_remote,
                is_iterator))
        if execute_remote:
            pass
            # # process.publish(self.get_topic() + "/_INPUT", count, True)
            # t = self.get_topic() + "/_START"
            # process.subscribe(t + '/ack', None)
            # process.publish(t, count, True)
            #
            # # wait for response
            # process.wait_topic(t + '/ack')
            # process.unsubscribe(t + '/ack', self)

        elif is_iterator:
            # node = self.nodes['NETWORK_CONDITION']
            node = self.condition_collector

            # for n in self.nodes['ALL_NETWORK_INPUT'].inputs.values() :
            for n in self.input_collector.inputs.values():
                n.force_count(count)

            # if self.process_name == process.process_name and process.is_mqtt_initialized():
            # pass #process.publish(self.get_topic() + "/_ITERATE", count,
            # True)

            internal_count = 0
            while True:
                logger.debug("Network.proceed: Loop {}".format(internal_count))

                # ループ終了判定
                self.proceed_frame(internal_count, process, outer_count=count)

                condition_satisfied = True
                if self.process_name == process.process_name:
                    logger.debug(
                        'Network.proceed: check if condition_satisfied: node={}'.format(
                            (type(node), node.name)))
                    condition_satisfied = node.get_output(
                        "OUTPUT", internal_count)
                    logger.debug(
                        'Network.proceed: condition_satisfied: {}'.format(condition_satisfied))

                if not condition_satisfied:
                    logger.debug("Network.proceed: breaking loop")
                    break

                # if last_count < 0 and not condition_satisfied:
                #     last_count = internal_count - 1

                # logger.debug("Network.proceed: count={}, last_count={}, input_count={}".format(internal_count, last_count, self.nodes['ALL_NETWORK_OUTPUT'].input_count))
                # lookAhead がある場合はもう何フレームが進めることになる。判定は出力段のフレーム番号で行う

                # if last_count >= 0:
                #     if self.output_collector.count >= last_count:
                #         logger.debug("Network.proceed: breaking loop")
                #         break

                internal_count += 1

            if self.process_name == process.process_name and process.is_mqtt_initialized():
                # process.publish(self.get_topic() + "/_ITERATE", count, False)
                pass

        else:
            count = self.proceed_frame(count, process, input=input)

        if count >= 0:
            # for (output_name, output) in
            # self.nodes['ALL_NETWORK_OUTPUT'].outputs.items() :
            for (output_name, output) in self.output_collectors.items():
                for output_proxy in output.outputs:
                    if output_proxy is not None:
                        output_proxy.push_data(count, process)
        return count

    def proceed_frame(self, count, process, outer_count=-1, input=None):
        logger.debug(
            "Network.proceed_frame: Proceed frame: {}, count={}, input={}".format(
                self.name, count, input))
        if outer_count < 0:
            outer_count = count

        logger.debug(
            'Network.proceed_frame: name={}, self.proess_name={}, process.process_name={}'.format(
                self.name,
                self.process_name,
                process.process_name))
        minimum_count = count
        if self.process_name == process.process_name:
            minimum_count = count
            for node in self.nodes.values():
                logger.debug("Network.proceed_frame: node: name={}, type={}, is-input={}, is-network={}, input-count={}".format(
                    node.name, node.type, (node.name[-6:] == "_input"), (isinstance(node, Network)), (node.get_input_count() == 0)))

            for node in self.nodes.values():
                if not isinstance(
                        node, Network) and node.name[-6:] != '_input' and node.get_input_count() == 0:
                    logger.debug(
                        "Network.proceed_frame: calling {} as spring node".format(
                            node.name))
                    minimum_count = min(
                        minimum_count, node.push_data(None, count, process))

            for node in self.nodes.values():
                if node.name[-6:] == '_input':
                    logger.debug(
                        "Network.proceed_frame: calling {} as network-input collector".format(node.name))

                    for iproxy in node.inputs.values():
                        if iproxy == input or input is None:
                            minimum_count = min(
                                minimum_count, node.push_data(
                                    iproxy, count, process, outer_count=outer_count))

            for node in self.nodes.values():
                if isinstance(node, Network) and len(node.inputs) == 0:
                    logger.debug(
                        "Network.proceed_frame: calling {} as spring network".format(
                            node.name))
                    logger.debug(
                        "Network.proceed_frame: node: {}, inputs={}, outputs={}".format(
                            node, node.inputs, node.outputs))
                    minimum_count = min(
                        minimum_count, node.push_data(None, count, process))

        else:
            while True:
                if process.to_quit:
                    logger.debug('process.to_quit: True')
                    break
                topic, count, data, args, _process = self.pop_mqtt_message()
                if topic is None:
                    process.check_mqtt()
                elif topic == self.get_topic() + "/_END":
                    break
                else:
                    self.process_mqtt_message(
                        topic, count, data, args, process)

        logger.debug(
            "Network.proceed_frame: Return from proceed frame: {}, count={}, minimum_count={}".format(
                self.name, count, minimum_count))
        return minimum_count

    def request(self, process):
        # [self.nodes['ALL_NETWORK_OUTPUT']]:
        for output in self.output_collectors.values():
            output.request(process)

    def initialize(self, process):
        logger.debug('Initialize: network: name={}, parent={}, process={}'.format(
            self.name, self.parent_network, (self.process_name, process.process_name)))
        node_counts = {}
        for node in self.nodes.values():
            logger.debug(
                'Initialize from network: 1: name={}'.format(node.name))
            node.initialize(process)
            node_counts[node] = 0
        if self.input_collector:
            # self.input_collector.initialize(process)
            node_counts[self.input_collector] = 0
        if self.output_collector:
            # self.output_collector.initialize(process)
            node_counts[self.output_collector] = 0
        if self.condition_collector:
            self.condition_collector.initialize(process)
            node_counts[self.condition_collector] = 0

        # for input in self.inputs.values():
        #     logger.debug(
        #         'Initialize from network: 2: name={}'.format(node.name))
        #     input.initialize(process)

        self.node_counts = node_counts  # ノードあたりの最新 count 数
        self.count = 0           # ネットワーク内の最大 count 数

        if self.parent_network:
            if self.parent_network.process_name != self.process_name and self.process_name == process.process_name:
                process.subscribe(self.get_topic() + '/_START', self, self)

    def wait_for_event(self, event):
        pass

    def release(self):
        for node in self.nodes.values():
            node.release()
        self.nodes = {}
        self.inputs = {}
        self.outputs = {}
        self.conditions = {}
        self.links = []
        self.count = 0

    def has_local_nodes(self, process):
        for n in self.nodes.values():
            if isinstance(n, Network):
                if n.has_local_nodes(process):
                    return True
            elif n.process_name == process.process_name:
                return True
        return False

    def has_remote_nodes(self, process):
        if self.process_name != process.process_name:
            return True
        for n in self.nodes.values():
            if isinstance(n, Network):
                if n.has_remote_nodes(process):
                    return True
            if n.process_name != process.process_name:
                return True
        return False

    def push_data(self, input, count, process):
        """ ループ count のデータが input に入力された場合にコールされる
        処理されたノードの集合を返す
        """
        return self.proceed(count, process, input=input)

    def pull_data(self, output, count):
        """
        """
        data = None
        for output in self.outputs.values():
            output.pull_data(self, count)

        return data

    def connect_nodes(self, node_from, terminal_from, node_to, terminal_to):
        node_from.connect_node(terminal_from, node_to, terminal_to)

    def get_input_count(self):
        return len(self.inputs)

    def get_output_count(self):
        return len(self.outputs)

    def get_topic(self):
        if self.topic is None:
            if self.network is not None:
                return "%s/%s" % (self.network.get_topic(), self.name)
            elif self.parent_network is not None:
                return "%s/%s" % (self.parent_network.get_topic(), self.name)
            else:
                return "%s" % self.name
        else:
            return self.topic

    def process_mqtt_message(self, topic, count, data, args, process):
        mytopic = self.get_topic()
        if topic.find(mytopic) != 0:
            return

        logger.debug(
            'Network.process_mqtt_message: topic={}, count={}, data={}, args={}, process={}'.format(
                topic,
                count,
                data,
                args,
                process))

        topic_tokens = topic[len(mytopic) + 1:].split('/')
        topic_tokens, action = topic_tokens[:-1], topic_tokens[-1]

        if len(topic_tokens) == 0 and action == '_START':
            # netork start event
            args.proceed(count, process)

        elif len(topic_tokens) == 1 and (action in ('_INPUT', '_START')):
            # node input event
            node = self.nodes[topic_tokens[-1]]
            if isinstance(node, Network):
                node.proceed(count, process)
            else:
                node.push_data(None, count, process)

        elif len(topic_tokens) == 2 and (action in ('_OUTPUT',)):
            # node input event
            node_name, output_name = topic_tokens
            node = self.nodes[node_name]
            for p in node.outputs[output_name]:
                p.push_data(count, data)

        elif topic == self.get_topic() + "/_END":
            pass

        elif topic == self.get_topic() + "/_ITERATE":
            pass

        elif topic == self.get_topic() + "/_CONDITION":
            pass

        else:
            raise BaseException('Unexpected message: topic={}'.format(topic))

    def add_mqtt_message(self, topic, count, data, args, process):
        self.mqtt_messages.append((topic, count, data, args, process))

    def pop_mqtt_message(self):
        if len(self.mqtt_messages) > 0:
            return self.mqtt_messages.pop()
        return (None, None, None, None, None)

#    def wait_mqtt_message(self, topic, count, process) :
#        while topic not in self.mqtt_messages or count not in self.mqtt_messages[topic] :
#            process.check_mqtt()
#        return self.mqtt_messages[topic].pop(count)

    def add_remote_data(self, count, proxy, args, data):
        while self.remote_data_offset > count or count >= self.remote_data_offset + \
                len(self.remote_data):
            self.remote_data.append([])

        self.remote_data[count -
                         self.remote_data_offset].append({'proxy': proxy, 'data': data})

    def get_remote_data(self, count):
        if self.remote_data_offset <= count and count < self.remote_data_offset + \
                len(self.remote_data):
            data = self.remote_data[count - self.remote_data_offset]
            self.remote_data[count - self.remote_data_offset] = []
            return data
        else:
            return []

    def has_remote_data(self, count):
        if self.remote_data_offset <= count and count < self.remote_data_offset + \
                len(self.remote_data):
            return len(self.remote_data[count - self.remote_data_offset]) > 0
        else:
            return False

    def dump(self):
        print('[Network]')
        print('  name: %s' % self.name)
        print('  type: %s' % self.type)

        print('  inputs: ')
        for (name, terminal) in self.inputs.items():
            print(
                '    Input: name=%s, node=%s, node-terminal=%s' %
                (name,
                 terminal.get_input_node().name if terminal.get_input_node() is not None else "(null)",
                 terminal.get_input_terminal_name()))

        print('  outputs: ')
        for (name, oo) in self.outputs.items():
            for output in oo:
                print(
                    '    Output: name=%s, node=%s, node-terminal=%s' %
                    (name,
                     terminal.get_output_node().name if terminal.get_output_node() is not None else "(null)",
                     terminal.get_output_terminal_name()))

        print('  conditions: ')
        for (name, terminal) in self.conditions.items():
            print('    Condition: name=%s, node=%s, node-terminal=%s' %
                  (name, terminal.node.name, terminal.name))

        print('  nodes: ')
        for (name, node) in self.nodes.items():
            node.dump()

    def __str__(self):
        return "<Network: type=%s, name=%s>" % (self.type, self.name)

# end of file
