/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#include "collector.h"

namespace HARK
{
    DECLARE_NODE(HARKMW_Collector);

    HARKMW_Collector::HARKMW_Collector(const std::string &nodeName, const ParameterSet &params) :
	BufferedNode(nodeName, params),
	last_input_count(0),
	last_output_count(0)
    {
    }

    ObjectRef
    HARKMW_Collector::getOutputNamed(const std::string &outputName, int count)
    {
      return BufferedNode::getOutputNamed(outputName, count);
	// for (unsigned int i = 0; i < this->inputs.size(); i++) {
	//     if(this->inputs[i].name == outputName) {
	// 	return this->getOutput(i, count);
	//     }
	// }
  //
	// throw new NodeException(this, std::string("Unknown output name :") + outputName, __FILE__, __LINE__);
    }

    void
    HARKMW_Collector::initialize()
    {
	this->BufferedNode::initialize();

  for(int i = 0; i < this->inputs.size(); i++){
    for(int j = 0; j < this->outputNames.size(); j++){
      if(this->inputs[i].name == this->outputNames[j]){
        this->out2inMap[j] = i;
      }
    }
  }

    }

    int
    HARKMW_Collector::translateInput(const std::string &inputName)
    {
	for (unsigned int i = 0; i < this->inputs.size(); i++) {
	    if(this->inputs[i].name == inputName) {
		return i;
	    }
	}

	return this->addInput(inputName);
    }

    int
    HARKMW_Collector::translateOutput(const std::string &outputName)
    {
      for (unsigned int i = 0; i < this->outputNames.size(); i++) {
    	    if(this->outputNames[i] == outputName) {
    		return i;
      }
    	}

    	return this->addOutput(outputName);
    }

    bool
    HARKMW_Collector::hasOutput(int output_id) const
    {
	return(int(inputs.size()) > output_id);
    }

    void
    HARKMW_Collector::request(int outputID, const ParameterSet &req)
    {
	this->inputs[outputID].node->request(this->inputs[outputID].outputID, req);
    }

    void
    HARKMW_Collector::requestAll(const ParameterSet &req)
    {
	for(int i = 0; i < this->inputs.size(); i++) {
	    this->inputs[i].node->request(this->inputs[i].outputID, req);
	}
    }

      ObjectRef
      HARKMW_Collector::getOutput(int output_id, int count)
      {
        return getOutput(output_id, this->last_input_count, count);
      }
  //   ObjectRef
  //   HARKMW_Collector::getOutput(int output_id, int count)
  //   {
  //     return BufferedNode::getOutput(output_id, count);
	// //std::cout << "HARKMW_Collector[" << this->name << ":" << (void *)this << "]::getOutput(" << output_id << ", " << count << ")" << std::endl;
  //
	// // if(!this->hasOutput(output_id)){
	// //     throw new NodeException(this,"Unknown output_id",__FILE__,__LINE__);
	// // }
  // //
	// // int target_count = this->last_input_count;
  // //
	// // try{
	// //     int outputID = this->inputs[output_id].outputID;
	// //     //std::cout << "input-count: " << this->last_input_count << ", output-count: " << this->last_output_count << ", target-count: " << target_count << std::endl;
	// //     ObjectRef r = (this->inputs[output_id].node)->getOutput(outputID, target_count);
	// //     return r;
  // //
	// // } catch (BaseException *e){
	// //     throw e->add(new NodeException (this, "Exception caught in Collector::getOutput", __FILE__, __LINE__));
	// // }
  //   }

    ObjectRef
    HARKMW_Collector::getOutputNamed(const std::string &outputName, int input_count, int output_count)
    {
	for (unsigned int i = 0; i < this->inputs.size(); i++) {
	    if(this->inputs[i].name == outputName) {
		return this->getOutput(i, input_count, output_count);
	    }
	}

	throw new NodeException(this, std::string("Unknown output name :") + outputName, __FILE__, __LINE__);
    }

    ObjectRef
    HARKMW_Collector::getOutput(int output_id, int input_count, int output_count)
    {
	// std::cout << "HARKMW_Collector[" << this->name << ":" << (void *)this << "]::getOutput(" << output_id << ", " << input_count << ", " << output_count << ")" << std::endl;

	if(!this->hasOutput(output_id)){
	    throw new NodeException(this,"Unknown output_id",__FILE__,__LINE__);
	}
	if(input_count < 0){
	    input_count = this->last_input_count;
	}
	if(output_count < 0){
	    output_count = this->last_output_count;
	}

	try{
	    // int outputID = this->inputs[output_id].outputID;
	    // //std::cout << "input-count: " << input_count << ", output-count: " << output_count << std::endl;
	    // ObjectRef r = (this->inputs[output_id].node)->getOutput(outputID, input_count);
      // std::cout << "Calling getOutput: " << output_id << ", " << input_count << std::endl;
      ObjectRef r = this->BufferedNode::getOutput(output_id, input_count);
	    this->last_input_count  = input_count;
	    this->last_output_count = output_count;
	    return r;

	} catch (BaseException *e){
	    throw e->add(new NodeException (this, "Exception caught in Collector::getOutput", __FILE__, __LINE__));
	}
    }

    void HARKMW_Collector::calculate(int output_id, int count, Buffer &out) {
      // std::cout << "HARKMW_Collector: calculate:  output_id=" << output_id << ", count=" << count << std::endl;

      int inputID = this->out2inMap[output_id];
      int outputID = this->inputs[inputID].outputID;
	    ObjectRef r = (this->inputs[inputID].node)->getOutput(outputID, count);
      out[count] = r;
    }

}

/* end of file */
