/* -*- mode: C++; coding: utf-8 -*- */
/*
 *
 *  HARK Middleware
 *  Copyright (C) 2013-2014 by Honda Research Institute Japan Co., Ltd.
 *
 *  $Id:$
 */

#include "libharkmw.h"
#include "Python.h"

#include <hark-base/Buffer.h>

#include "proxy.h"
#include "node.h"
#include "object.h"

#ifdef _WIN64
#include <windows.h>
#include <errno.h>

#define dlopen(P,G) (void*)LoadLibrary(P)
__inline const char* dlerror()
{
    static char szMsgBuf[256];
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                  NULL,
                  GetLastError(),
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                  szMsgBuf,
                  sizeof szMsgBuf,
                  NULL);
    return szMsgBuf;
}
#else
#include <dlfcn.h>
#endif

namespace HARKMW
{
    static PyObject *
    load_dynamic(PyObject *self, PyObject *args)
    {
        const char *path;
        if(!PyArg_ParseTuple(args, "s", &path)){
            return NULL;
        }

        if(dlopen(path, RTLD_LAZY | RTLD_GLOBAL) == NULL){
            PyErr_SetString(PyExc_RuntimeError, dlerror());
            return NULL;
        }else{
            Py_RETURN_NONE;
        }
    }

    static void
    destruct_fd_buffer(PyObject *p)
    {
        delete (HARK::Buffer *)p;
    }

    static PyObject *
    create_buffer(PyObject *self, PyObject *args)
    {
        int c;
        if(!PyArg_ParseTuple(args, "i", &c)){
            return NULL;
        }

        return PyCapsule_New(new HARK::Buffer(c), NULL,destruct_fd_buffer);
    }

    static PyObject *
    print_modules(PyObject *_self, PyObject *args)
    {
        std::map<std::string, HARK::ModuleInfo>::iterator it = HARK::Object::module_info_dic.begin();
        while (it != HARK::Object::module_info_dic.end()) {
            std::stringstream ss;
            ss << it->first << " " << it->second.module_version;
            if (it->second.svn_version != "") {
                ss << " (Revision:" << it->second.svn_version << ")";
            }
            std::cout << ss.str() << std::endl;
            it++;
        }
        Py_RETURN_NONE;
    }

    class ObjectRefWrap
    {
    private:
        HARK::ObjectRef objectRef;

    public:
        ObjectRefWrap(HARK::ObjectRef &ref){
            objectRef = ref;
        }

    };
}

static PyMethodDef harkmw_methods[] = {
    {"load_dynamic", HARKMW::load_dynamic, METH_VARARGS, "Load HARK's type library"},
    {"create_buffer", HARKMW::create_buffer, METH_VARARGS, "Create native HARK::Buffer object"},
    {"print_modules", HARKMW::print_modules, METH_NOARGS, "Print versions of loaded modules"},
    {NULL, NULL, 0, NULL},
};

static struct PyModuleDef moduledef = {
        PyModuleDef_HEAD_INIT,
        "harkmwnative",
        NULL,
        -1,
        harkmw_methods
};

MODULE_INIT_FUNC(harkmwnative)
{
    PyObject *m = NULL;

    if(PyType_Ready(&HARKMW::NodeType) < 0){
        return NULL;
    }

    if(PyType_Ready(&HARKMW::ProxyType) < 0){
        return NULL;
    }

    HARKMW::ObjectType.tp_new = PyType_GenericNew;
    if(PyType_Ready(&HARKMW::ObjectType) < 0){
        return NULL;
    }

    m = PyModule_Create(&moduledef);

    Py_INCREF(&HARKMW::NodeType);
    PyModule_AddObject(m, "Node", (PyObject *)&HARKMW::NodeType);

    Py_INCREF(&HARKMW::ProxyType);
    PyModule_AddObject(m, "Proxy", (PyObject *)&HARKMW::ProxyType);

    Py_INCREF(&HARKMW::ObjectType);
    PyModule_AddObject(m, "Object", (PyObject *)&HARKMW::ObjectType);

    return m;
}

int test(PyObject *obj)
{
    return 0;
}
